/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <memory>
#include <numeric>
#include <string>

// ArmarX
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/core/data/variant/complex/NDArray.h>

namespace armarx::aron::data::converter
{
    class AronVectorConverter
    {
    private:
        template <typename T>
        static std::vector<T>
        convert_to_1d_vector(const unsigned char* data, const int offset, const int size)
        {
            ARMARX_CHECK_NOT_NULL(data);
            ARMARX_CHECK(size % sizeof(T) == 0);
            int elements = size / sizeof(T);

            std::vector<T> v(elements);
            memcpy(v.data(), data + offset, size);

            return v;
        }

    public:
        AronVectorConverter() = delete;

        template <typename T>
        static std::vector<std::vector<T>>
        ConvertTo2DVector(const data::NDArray& nav, int dim0size = -1)
        {
            const auto& dims = nav.getShape();

            if (dims.size() == 0 || dim0size == 0)
            {
                return {};
            }

            if (dim0size == -1)
            {
                dim0size = dims.at(0);
            }

            const int fullDataByteSize = nav.getDataAsVector().size();

            ARMARX_CHECK(fullDataByteSize % sizeof(T) == 0);
            const int fullDataSize = fullDataByteSize / sizeof(T);

            ARMARX_CHECK(fullDataSize % dim0size == 0);
            const int dim1size = fullDataSize / dim0size;

            const int oneRowByteSize = dim1size * sizeof(T);
            std::vector<std::vector<T>> v(dims.at(0));
            for (int i = 0; i < dim0size; i++)
            {
                v[i] = convert_to_1d_vector<T>(nav.getData(), i * oneRowByteSize, oneRowByteSize);
            }

            return v;
        }

        template <typename T>
        static std::vector<std::vector<T>>
        ConvertTo2DVector(const data::NDArrayPtr& nav)
        {
            ARMARX_CHECK_NOT_NULL(nav);

            return ConvertTo2DVector<T>(*nav);
        }

        template <typename T>
        static std::vector<T>
        ConvertTo1DVector(const data::NDArray& nav)
        {
            const auto& dims = nav.getShape();

            if (dims.size() == 0)
            {
                return {};
            }

            const int fullDataByteSize = nav.getDataAsVector().size();

            ARMARX_CHECK(fullDataByteSize % sizeof(T) == 0);
            const int dim0size = fullDataByteSize / sizeof(T);

            const int oneRowByteSize = dim0size * sizeof(T);
            std::vector<T> v = convert_to_1d_vector<T>(nav.getData(), 0, oneRowByteSize);
            return v;
        }

        template <typename T>
        static std::vector<T>
        ConvertTo1DVector(const data::NDArrayPtr& nav)
        {
            ARMARX_CHECK_NOT_NULL(nav);

            return ConvertTo1DVector<T>(*nav);
        }

        // alias
        template <typename T>
        static std::vector<T>
        ConvertToVector(const data::NDArrayPtr& nav)
        {
            return ConvertTo1DVector<T>(nav);
        }

        // Attention: If a vector was flattened, the reconstruction is flattened as well!
        template <typename T>
        static data::NDArrayPtr
        ConvertFrom1DVector(const std::vector<T>& data)
        {
            data::NDArrayPtr ndArr(new data::NDArray);

            ndArr->setShape({static_cast<int>(data.size()), sizeof(T)});
            ndArr->setData(sizeof(T) * data.size(),
                           reinterpret_cast<const unsigned char*>(data.data()));

            return ndArr;
        }

        // alias
        template <typename T>
        static data::NDArrayPtr
        ConvertFromVector(const std::vector<T>& data)
        {
            return ConvertFrom1DVector<T>(data);
        }
    };
} // namespace armarx::aron::data::converter

// legacy
namespace armarx::aron::converter
{
    using AronVectorConverter = ::armarx::aron::data::converter::AronVectorConverter;
}
