#pragma once

// std/stl
#include <functional>
#include <mutex>
#include <queue>
#include <thread>

// base class
#include <ArmarXCore/core/logging/Logging.h>

// ArmarX
#include <ArmarXCore/core/time/DateTime.h>
#include <ArmarXCore/core/time/Metronome.h>

#include <RobotAPI/interface/skills/SkillManagerInterface.h>
#include <RobotAPI/libraries/aron/core/data/variant/All.h>

#include "SkillDescription.h"
#include "SkillID.h"
#include "SkillPreparationInput.h"
#include "SkillStatusUpdate.h"
#include "error/Exception.h"

namespace armarx
{
    namespace skills
    {
        class Skill : public armarx::Logging
        {
        public:
            using CallbackT =
                std::function<void(const SkillStatus s, const armarx::aron::data::DictPtr&)>;

            struct InitResult
            {
                TerminatedSkillStatus status;
            };

            struct PrepareResult
            {
                ActiveOrTerminatedSkillStatus status;
            };

            struct MainResult
            {
                TerminatedSkillStatus status;
                aron::data::DictPtr data = nullptr;
            };

            struct ExitResult
            {
                TerminatedSkillStatus status;
            };

            Skill() = delete;
            Skill(const SkillDescription&);
            virtual ~Skill() = default;

            /// The id of the skill (combination of provider and name must be unique).
            SkillID
            getSkillId() const
            {
                return description.skillId;
            }

            SkillDescription
            getSkillDescription() const
            {
                return description;
            }

            void
            setProviderId(const skills::ProviderID& pid)
            {
                description.skillId.providerId = pid;
            }

            PrepareResult prepareSkill();

            InitResult initSkill();

            MainResult mainOfSkill();

            ExitResult exitSkill();

            // Condition listeners
            // used to notify the skill from extern to stop
            void notifySkillToStopASAP();

            // returns whether the skill should terminate as soon as possible
            bool checkWhetherSkillShouldStopASAP() const;

            // merge parameters to the local parameters of the skill
            void
            addParameters(const aron::data::DictPtr& d)
            {
                std::scoped_lock l(this->parametersMutex);
                if (this->parameters == nullptr)
                {
                    this->parameters = d;
                }
                else
                {
                    this->parameters->mergeAndReplaceCopy(d);
                }
            }

            // hard set the parameters, ignoring everything that has been set or merged before
            void
            setParameters(const aron::data::DictPtr& d)
            {
                std::scoped_lock l(this->parametersMutex);
                this->parameters = d;
            }

            // get the parameters
            aron::data::DictPtr
            getParameters() const
            {
                std::scoped_lock l(this->parametersMutex);
                return this->parameters;
            }

        protected:
            static MainResult MakeAbortedResult(aron::data::DictPtr data = nullptr);

            // fires if the skill reaches timeout
            void notifyTimeoutReached();

            // helper methods to do all the static initialization stuff
            InitResult _init();
            PrepareResult _prepare();
            MainResult _main();
            ExitResult _exit();

            /// Override this method with the actual implementation.
            virtual InitResult init();

            /// Override this method with the actual implementation.
            virtual PrepareResult prepare();

            /// Override this method with the actual implementation. The callback is for status updates to the calling instance
            virtual MainResult main();

            /// Override this method with the actual implementation.
            virtual ExitResult exit();

        private:
            /// Override these methods if you want to do something special when notification comes
            virtual void onTimeoutReached();
            virtual void onStopRequested();

        protected:
            /// install a condition which is frequently checked from the conditionCheckingThread
            void installConditionWithCallback(std::function<bool()>&& f,
                                              std::function<void()>&& cb);

        public:
            // running params
            armarx::core::time::DateTime started = armarx::core::time::DateTime::Invalid();
            armarx::core::time::DateTime exited = armarx::core::time::DateTime::Invalid();

            // parameterization. Will be set from implementation wrapper.
            // const params after initialization!!
            CallbackT callback;
            manager::dti::SkillManagerInterfacePrx manager = nullptr;
            std::string executorName = "INVALID EXECUTOR NAME";

        protected:
            // non-const params
            mutable std::mutex parametersMutex;
            armarx::aron::data::DictPtr parameters = nullptr;

            // The descripion of the skill
            SkillDescription description;

            // active conditions. First is condition (bool return func)
            mutable std::mutex conditionCallbacksMutex;
            std::vector<std::pair<std::function<bool()>, std::function<void()>>>
                conditionCallbacks = {};


            /// Use conditions this way
            std::atomic_bool initializing = false;
            std::atomic_bool preparing = false;
            std::atomic_bool running = false;
            std::atomic_bool stopped = false;
            std::atomic_bool timeoutReached = false;


        private:
            std::thread conditionCheckingThread; // A therad that checks the conditions frequently
            armarx::Frequency conditionCheckingThreadFrequency = armarx::Frequency::Hertz(20);
        };

        template <class T>
        concept isSkill = std::is_base_of<Skill, T>::value;

    } // namespace skills
} // namespace armarx
