#include "SkillProviderComponentPlugin.h"

#include <ArmarXCore/core/Component.h>

#include <RobotAPI/libraries/aron/core/data/variant/primitive/All.h>
#include <RobotAPI/libraries/skills/core/ProviderInfo.h>

namespace armarx::plugins
{
    void
    SkillProviderComponentPlugin::preOnInitComponent()
    {
    }

    void
    SkillProviderComponentPlugin::preOnConnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        p.getProxy(myPrx, -1);
    }

    void
    SkillProviderComponentPlugin::postOnConnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        const std::string providerName = p.getName();

        // register self to manager
        skills::ProviderInfo i(skills::ProviderID(providerName), myPrx, getSkillDescriptions());

        ARMARX_INFO << "Adding provider to manager: " << i.providerId;
        manager->addProvider(i.toIce()); // add provider info to manager
    }

    void
    SkillProviderComponentPlugin::preOnDisconnectComponent()
    {
        auto& p = parent<SkillProviderComponentPluginUser>();
        std::string providerName = p.getName();

        auto id = skills::manager::dto::ProviderID{providerName};
        manager->removeProvider(id);
    }

    void
    SkillProviderComponentPlugin::postCreatePropertyDefinitions(PropertyDefinitionsPtr& properties)
    {
        std::string prefix = "skill.";
        properties->component(
            manager,
            "SkillMemory",
            prefix + "SkillManager",
            "The name of the SkillManager (or SkillMemory) proxy this provider belongs to.");
    }

    void
    SkillProviderComponentPlugin::addSkillFactory(std::unique_ptr<skills::SkillFactory>&& fac)
    {
        if (!fac)
        {
            return;
        }

        auto& p = parent<SkillProviderComponentPluginUser>();
        const std::string componentName = p.getName();

        const skills::ProviderID providerId({componentName});

        // lock skills map
        const std::unique_lock l(skillFactoriesMutex);
        auto skillId = fac->createSkillDescription(providerId).skillId;

        if (skillFactories.find(skillId) != skillFactories.end())
        {
            ARMARX_WARNING << "Try to add a skill factory for skill '" + skillId.toString() +
                                  "' which already exists in list. Ignoring this skill.";
            return;
        }

        ARMARX_INFO << "Adding skill `" << skillId << "` to component `" << componentName << "` .";

        skillFactories.emplace(skillId, std::move(fac));

        //        if (connected)
        //        {
        //            // if skill is added after onConnect we have to set the proxies manually.
        //            std::string providerName = parent().getName();
        //            s.first->second.skill->manager = manager;
        //            s.first->second.skill->providerName = providerName;
        //        }
    }

    skills::SkillFactory*
    SkillProviderComponentPlugin::addSkillFactory(const skills::SkillDescription& desc,
                                                  const skills::LambdaSkill::FunT& f)
    {
        return addSkillFactory<skills::LambdaSkill>(desc, f);
    }

    skills::SkillFactory*
    SkillProviderComponentPlugin::getSkillFactory(const armarx::skills::SkillID& skillId)
    {
        // NON BLOCKING: WE ASSERT THAT THE LOCK IS ALREADY TAKEN
        ARMARX_CHECK(skillId.fullySpecified());

        if (skillFactories.count(skillId) == 0)
        {
            ARMARX_INFO << "Could not find a skill factory for id: " << skillId;
            return nullptr;
        }

        auto* facPtr = skillFactories.at(skillId).get();
        return static_cast<skills::SkillFactory*>(facPtr);
    }

    std::optional<skills::SkillStatusUpdate>
    SkillProviderComponentPlugin::getSkillExecutionStatus(
        const skills::SkillExecutionID& execId) const
    {
        ARMARX_CHECK(execId.skillId.skillSpecified());

        const std::unique_lock l(skillExecutionsMutex);
        if (skillExecutions.find(execId) != skillExecutions.end())
        {
            ARMARX_WARNING << "Skill execution for skill '" + execId.skillId.toString() +
                                  "' not found!";
            return std::nullopt;
        }

        return skillExecutions.at(execId).statusUpdate;
    }

    std::map<skills::SkillExecutionID, skills::SkillStatusUpdate>
    SkillProviderComponentPlugin::getSkillExecutionStatuses() const
    {
        std::map<skills::SkillExecutionID, skills::SkillStatusUpdate> skillUpdates;

        const std::scoped_lock l(skillExecutionsMutex);
        for (const auto& [key, impl] : skillExecutions)
        {
            const std::scoped_lock l2(impl.skillStatusesMutex);
            skillUpdates.insert({key, impl.statusUpdate});
        }
        return skillUpdates;
    }

    std::optional<skills::SkillDescription>
    SkillProviderComponentPlugin::getSkillDescription(const skills::SkillID& skillId) const
    {
        ARMARX_CHECK(skillId.fullySpecified());

        const std::unique_lock l(skillFactoriesMutex);
        if (skillFactories.find(skillId) != skillFactories.end())
        {
            ARMARX_WARNING << "Skill description for skill '" + skillId.toString() + "' not found!";
            return std::nullopt;
        }

        return skillFactories.at(skillId)->createSkillDescription(*skillId.providerId);
    }

    std::map<skills::SkillID, skills::SkillDescription>
    SkillProviderComponentPlugin::getSkillDescriptions() const
    {
        std::map<skills::SkillID, skills::SkillDescription> skillDesciptions;
        const std::unique_lock l(skillFactoriesMutex);
        for (const auto& [key, fac] : skillFactories)
        {
            ARMARX_CHECK(key.fullySpecified());
            skillDesciptions.insert({key, fac->createSkillDescription(*key.providerId)});
        }
        return skillDesciptions;
    }

    skills::SkillStatusUpdate
    SkillProviderComponentPlugin::executeSkill(
        const skills::SkillExecutionRequest& executionRequest)
    {
        ARMARX_CHECK(executionRequest.skillId.fullySpecified());

        // The skill will be executed in a seperate thread
        std::thread execution;

        // setup input args for skill execution
        skills::SkillParameterization usedParameterization{executionRequest.params,
                                                           executionRequest.callbackInterface};

        skills::SkillExecutionID executionId(executionRequest.skillId,
                                             executionRequest.executorName,
                                             armarx::core::time::DateTime::Now());

        skills::SkillStatusUpdate ret{executionId, usedParameterization};
        {
            auto l1 = std::unique_lock{skillFactoriesMutex};

            const auto& fac = getSkillFactory(executionRequest.skillId);
            ARMARX_CHECK(fac) << "Could not find a factory for skill " << executionRequest.skillId;

            // async start execution. But we wait for the execution to finish at the end of this method
            execution = std::thread(
                [&]()
                {
                    skills::detail::SkillImplementationWrapper* wrapper;
                    {
                        const std::unique_lock l2{skillExecutionsMutex};
                        auto it = skillExecutions.emplace(
                            std::piecewise_construct,
                            std::make_tuple(executionId),
                            std::make_tuple(*fac, executionId, usedParameterization));

                        ARMARX_CHECK(it.second)
                            << "For some reason a skill '" << executionId.skillId.toString()
                            << "' execution failed when instantiating the "
                               "wrapper class and adding it to the internally used map. This "
                               "should "
                               "not happen!!";

                        wrapper = &it.first->second;
                    }


                    // execute waits until the previous execution finishes.
                    auto x = wrapper->executeSkill();
                    ret.data = x.data;
                    ret.status = armarx::skills::toSkillStatus(x.status);
                });
        } // release lock. We don't know how long the skill needs to finish and we have to release the lock for being able to abort the execution

        if (execution.joinable())
        {
            execution.join();

            // tidy up map of executions
            const std::unique_lock l2{skillExecutionsMutex};
            if (auto it = skillExecutions.find(executionId); it != skillExecutions.end())
            {
                skillExecutions.erase(it);
            }
        }
        return ret;
    }

    skills::SkillExecutionID
    SkillProviderComponentPlugin::executeSkillAsync(
        const skills::SkillExecutionRequest& executionRequest)
    {
        ARMARX_CHECK(executionRequest.skillId.fullySpecified());

        // The skill will be executed in a seperate thread
        std::thread execution;

        // setup input args for skill execution
        skills::SkillParameterization usedParameterization{executionRequest.params,
                                                           executionRequest.callbackInterface};

        skills::SkillExecutionID executionId(executionRequest.skillId,
                                             executionRequest.executorName,
                                             armarx::core::time::DateTime::Now());

        {
            auto l1 = std::unique_lock{skillFactoriesMutex};

            const auto& fac = getSkillFactory(executionRequest.skillId);
            ARMARX_CHECK(fac) << "Could not find a factory for skill " << executionRequest.skillId;

            // async start execution. But we wait for the execution to finish at the end of this method
            execution = std::thread(
                [&]()
                {
                    skills::detail::SkillImplementationWrapper* wrapper;
                    {
                        const std::unique_lock l2{skillExecutionsMutex};
                        auto it = skillExecutions.emplace(
                            std::piecewise_construct,
                            std::make_tuple(executionId),
                            std::make_tuple(*fac, executionId, usedParameterization));

                        ARMARX_CHECK(it.second)
                            << "For some reason a skill '" << executionId.skillId.toString()
                            << "' execution failed when instantiating the "
                               "wrapper class and adding it to the internally used map. This "
                               "should "
                               "not happen!!";

                        wrapper = &it.first->second;
                    }


                    // execute waits until the previous execution finishes.
                    auto x = wrapper->executeSkill();
                });
        } // release lock. We don't know how long the skill needs to finish and we have to release the lock for being able to abort the execution

        execution.detach();
        return executionId;
    }

    void
    SkillProviderComponentPlugin::addSkillParameters(const skills::SkillExecutionID& executionId,
                                                     const armarx::aron::data::DictPtr& input)
    {
        ARMARX_CHECK(executionId.skillId.fullySpecified());

        const std::scoped_lock l{skillExecutionsMutex};
        auto it = skillExecutions.find(executionId);
        if (it == skillExecutions.end())
        {
            ARMARX_INFO << "No acive execution for skill '" + executionId.skillId.toString() +
                               "' found! Ignoring prepareSkill request.";
            return;
        }

        std::scoped_lock l2{it->second.skillStatusesMutex};
        if (it->second.statusUpdate.status != skills::SkillStatus::Preparing)
        {
            ARMARX_INFO << "Could not prepare the skill '" + executionId.skillId.toString() +
                               "' because its not in preparing phase.";
            return;
        }

        it->second.addSkillParameters(input);
    }

    void
    SkillProviderComponentPlugin::abortSkill(const skills::SkillExecutionID& executionId)
    {
        ARMARX_CHECK(executionId.skillId.fullySpecified());

        const std::unique_lock l(skillExecutionsMutex);
        auto it = skillExecutions.find(executionId);
        if (it == skillExecutions.end())
        {
            ARMARX_INFO << "No acive execution for skill '" + executionId.skillId.toString() +
                               "' found! Ignoring abortSkill request.";
            return;
        }

        it->second.stopSkill();
    }

} // namespace armarx::plugins

namespace armarx
{
    SkillProviderComponentPluginUser::SkillProviderComponentPluginUser()
    {
        addPlugin(plugin);
    }

    IceUtil::Optional<skills::provider::dto::SkillDescription>
    SkillProviderComponentPluginUser::getSkillDescription(
        const skills::provider::dto::SkillID& skillId,
        const Ice::Current& /*unused*/)
    {
        auto id = skills::SkillID::FromIce(skillId, getName());
        auto o = plugin->getSkillDescription(id);
        if (o.has_value())
        {
            return o->toProviderIce();
        }
        return {};
    }

    skills::provider::dto::SkillDescriptionMap
    SkillProviderComponentPluginUser::getSkillDescriptions(const Ice::Current& /*unused*/)
    {
        skills::provider::dto::SkillDescriptionMap ret;
        for (const auto& [k, v] : plugin->getSkillDescriptions())
        {
            ret.insert({k.toProviderIce(), v.toProviderIce()});
        }
        return ret;
    }

    IceUtil::Optional<skills::provider::dto::SkillStatusUpdate>
    SkillProviderComponentPluginUser::getSkillExecutionStatus(
        const skills::provider::dto::SkillExecutionID& executionId,
        const Ice::Current& /*unused*/)
    {
        auto execId = skills::SkillExecutionID::FromIce(executionId, getName());
        auto o = plugin->getSkillExecutionStatus(execId);
        if (o.has_value())
        {
            return o->toProviderIce();
        }
        return {};
    }

    skills::provider::dto::SkillStatusUpdateMap
    SkillProviderComponentPluginUser::getSkillExecutionStatuses(const Ice::Current& /*unused*/)
    {
        skills::provider::dto::SkillStatusUpdateMap ret;
        for (const auto& [k, v] : plugin->getSkillExecutionStatuses())
        {
            ret.insert({k.toProviderIce(), v.toProviderIce()});
        }
        return ret;
    }

    // Please not that this method waits until the skill can be scheduled!
    skills::provider::dto::SkillStatusUpdate
    SkillProviderComponentPluginUser::executeSkill(
        const skills::provider::dto::SkillExecutionRequest& info,
        const Ice::Current& /*unused*/)
    {
        auto exec = skills::SkillExecutionRequest::FromIce(info, getName());
        auto up = this->plugin->executeSkill(exec);
        return up.toProviderIce();
    }

    skills::provider::dto::SkillExecutionID
    SkillProviderComponentPluginUser::executeSkillAsync(
        const skills::provider::dto::SkillExecutionRequest& info,
        const Ice::Current& current /*unused*/)
    {
        auto exec = skills::SkillExecutionRequest::FromIce(info, getName());
        auto id = this->plugin->executeSkillAsync(exec);
        return id.toProviderIce();
    }

    void
    SkillProviderComponentPluginUser::addSkillParameters(
        const skills::provider::dto::SkillExecutionID& id,
        const aron::data::dto::DictPtr& input,
        const Ice::Current& current /*unused*/)
    {
        auto exec = skills::SkillExecutionID::FromIce(id, getName());
        auto prep = armarx::aron::data::Dict::FromAronDictDTO(input);
        this->plugin->addSkillParameters(exec, prep);
    }

    void
    SkillProviderComponentPluginUser::abortSkill(const skills::provider::dto::SkillExecutionID& id,
                                                 const Ice::Current& /*unused*/)
    {
        auto exec = skills::SkillExecutionID::FromIce(id, getName());
        this->plugin->abortSkill(exec);
    }

    const std::experimental::observer_ptr<plugins::SkillProviderComponentPlugin>&
    SkillProviderComponentPluginUser::getSkillProviderPlugin() const
    {
        return plugin;
    }
} // namespace armarx
