#include "SkillImplementationWrapper.h"

namespace armarx
{
    namespace skills::detail
    {
        SkillImplementationWrapper::SkillImplementationWrapper(
            const skills::SkillFactory& fac,
            const skills::SkillExecutionID& execId,
            const skills::SkillParameterization& param) :
            factory(fac), statusUpdate(execId, param)
        {
        }

        // ask a skill to stop
        void
        SkillImplementationWrapper::stopSkill()
        {
            while (this->skill == nullptr)
            {
                ARMARX_INFO << deactivateSpam()
                            << "Trying to stop a skill that has not been constructed yet... "
                               "Waiting until construction...";
                std::this_thread::sleep_for(std::chrono::milliseconds(100));
            }

            this->skill->notifySkillToStopASAP();
        }

        void
        SkillImplementationWrapper::addSkillParameters(const aron::data::DictPtr& i)
        {
            while (this->skill == nullptr)
            {
                ARMARX_INFO << deactivateSpam()
                            << "Trying to add params to a skill that has not been constructed "
                               "yet... Waiting until construction...";
                std::this_thread::sleep_for(std::chrono::milliseconds(100));
            }

            this->skill->addParameters(i);
        }

        TerminatedSkillStatusUpdate
        SkillImplementationWrapper::executeSkill()
        {
            std::scoped_lock l(this->executionMutex);
            // -------------------------------------------------------------------------------------
            // sanity check
            // -------------------------------------------------------------------------------------
            ARMARX_CHECK(this->skill == nullptr)
                << "A SKILL HAS ALREADY BEEN CONSTRUCTED. THIS SHOULD NOT HAPPEN!!";


            // -------------------------------------------------------------------------------------
            // setup basic vars and lambdas
            // -------------------------------------------------------------------------------------
            // actually we should lock... however this is only read access and the members are const throughout the execution...
            ARMARX_CHECK(statusUpdate.executionId.skillId.fullySpecified());

            const auto& initial_aron_params = statusUpdate.usedParameterization.parameterization;
            const auto& callback_interface = statusUpdate.usedParameterization.callbackInterface;
            const auto& skillId = statusUpdate.executionId.skillId;
            const auto& skillName = skillId.skillName;
            const auto& providerId = *skillId.providerId;
            const auto& executorName = statusUpdate.executionId.executorName;

            ARMARX_INFO_S << "Executing skill: " << skillName;

            auto updateStatus =
                [&](const SkillStatus status, const aron::data::DictPtr& data = nullptr)
            {
                std::unique_lock l(skillStatusesMutex);
                statusUpdate.status = status;
                statusUpdate.data = data;
                if (skill) // if skill has been constructed
                {
                    // update parameterization
                    statusUpdate.usedParameterization.parameterization = skill->getParameters();
                }

                if (callback_interface) // if callback interface is used
                {
                    auto pid = providerId.toCallbackIce();
                    callback_interface->updateStatusForSkill(statusUpdate.toProviderIce(), pid);
                }
            };

            auto createErrorMessageData = [](const std::string& message)
            {
                auto obj = aron::make_dict();
                auto m = aron::make_string(message, aron::Path({"errormessage"}));
                obj->addElement("errormessage", m);
                return obj;
            };

            {
                std::unique_lock l(skillStatusesMutex);

                // ---------------------------------------------------------------------------------
                // check times
                // ---------------------------------------------------------------------------------
                if ((armarx::core::time::DateTime::Now() -
                     statusUpdate.executionId.executionStartedTime) >
                    armarx::core::time::Duration::Seconds(1))
                {
                    ARMARX_WARNING_S << "SkillError 001: For some reason the skill '"
                                     << skillId.toString()
                                     << "' has been scheduled > 1s ago. The execution should start "
                                        "much faster! Continue, but this is bad behavior..";
                }
            }

            // -------------------------------------------------------------------------------------
            // construct skill
            // -------------------------------------------------------------------------------------
            updateStatus(SkillStatus::Constructing);
            this->skill = this->factory.createSkill(providerId);
            this->skill->executorName = (executorName);
            this->skill->manager = skills::manager::dti::SkillManagerInterfacePrx::checkedCast(
                callback_interface); // ugly. Get managerPrx from manager!
            ;
            this->skill->callback = [&](const SkillStatus s, const armarx::aron::data::DictPtr& d)
            { updateStatus(s, d); };

            // set initial parameters that were attached to the execution request (only add as we are not sure whether some updates already arrived)
            skill->addParameters(initial_aron_params);


            auto makeTerminationResult = [&](const std::string& message)
            {
                updateStatus(SkillStatus::Failed, createErrorMessageData(message));

                std::unique_lock l(skillStatusesMutex);
                auto terminated = TerminatedSkillStatusUpdate(statusUpdate.executionId,
                                                              statusUpdate.usedParameterization);
                terminated.data = statusUpdate.data;
                terminated.status = TerminatedSkillStatus::Failed;
                return terminated;
            };

            // -------------------------------------------------------------------------------------
            // Check params
            // -------------------------------------------------------------------------------------
            // NOT RELEVANT ANYMORE!!
            //            if (skill->getSkillDescription().acceptedType && initial_aron_params &&
            //                not(initial_aron_params->fullfillsType(skill->getSkillDescription().acceptedType)))
            //            {
            //                std::string message =
            //                    "SkillError 002: The Skill '" + skillName +
            //                    "' has a type and got parameters but the input does not match the type.";
            //                ARMARX_ERROR_S << message;
            //                return makeTerminationResult(message);
            //            }


            auto exitAndMakeTerminationResult = [&](const std::string& message)
            {
                skill->exitSkill(); // try to exit skill. Ignore return value
                return makeTerminationResult(message);
            };

            // -------------------------------------------------------------------------------------
            // Init skill
            // -------------------------------------------------------------------------------------
            try
            {
                updateStatus(SkillStatus::Initializing);
                Skill::InitResult initRet = skill->initSkill();
                if (initRet.status != TerminatedSkillStatus::Succeeded)
                {
                    std::string message = "SkillError 101: The initialization of skill '" +
                                          skillName + "' did not succeed.";
                    return exitAndMakeTerminationResult(message);
                }
            }

            catch (const std::exception& ex)
            {
                std::string message =
                    "SkillError 101e: An error occured during the initialization of skill '" +
                    skillName + "'. The error was: " + GetHandledExceptionString();
                ARMARX_ERROR_S << message;
                return exitAndMakeTerminationResult(message);
            }

            // -------------------------------------------------------------------------------------
            // Prepare skill
            // -------------------------------------------------------------------------------------
            try
            {
                updateStatus(SkillStatus::Preparing);
                auto prepareRet = skill->prepareSkill();
                while (prepareRet.status == ActiveOrTerminatedSkillStatus::Running)
                {
                    ARMARX_INFO << deactivateSpam() << "Not all requirements for skill "
                                << skillName << " fulfilled. Waiting...";

                    // wait...
                    prepareRet = skill->prepareSkill();
                    std::this_thread::sleep_for(std::chrono::milliseconds(50));
                }

                if (prepareRet.status != ActiveOrTerminatedSkillStatus::Succeeded)
                {
                    std::string message = "SkillError 201: The prepare method of skill '" +
                                          skillName + "' did not succeed.";
                    return exitAndMakeTerminationResult(message);
                }
            }

            catch (const std::exception& ex)
            {
                std::string message = "SkillError 201e: An error occured during waiting for skill "
                                      "dependencies of skill '" +
                                      skillName +
                                      "'. The error was: " + GetHandledExceptionString();
                ARMARX_ERROR_S << message;
                return exitAndMakeTerminationResult(message);
            }

            // execute. If the skill fails for some reason, from this point it will always execute its exit function.
            updateStatus(SkillStatus::Running);

            // Init succeeded!


            Skill::MainResult mainRet;
            try
            {
                mainRet = skill->mainOfSkill();
                if (mainRet.status != TerminatedSkillStatus::Succeeded)
                {
                    std::string message = "SkillError 501: The main method of skill '" + skillName +
                                          "' did not succeed.";
                    return exitAndMakeTerminationResult(message);
                }
            }

            catch (const std::exception& ex)
            {
                std::string message =
                    "SkillError 501e: An error occured during the main method of skill '" +
                    skillName + "'. The error was: " + GetHandledExceptionString();
                ARMARX_ERROR_S << message;
                return exitAndMakeTerminationResult(message);
            }

            // Main succeeded!

            try
            {
                Skill::ExitResult exitRet = skill->exitSkill();
                if (exitRet.status != TerminatedSkillStatus::Succeeded)
                {
                    std::string message = "SkillError 601: The exit method of skill '" + skillName +
                                          "' did not succeed.";
                    return makeTerminationResult(message);
                }
            }

            catch (const std::exception& ex)
            {
                std::string message =
                    "SkillError 601e: An error occured during the exit method of skill '" +
                    skillName + "'. The error was: " + GetHandledExceptionString();
                ARMARX_ERROR_S << message;
                return makeTerminationResult(message);
            }

            // Exit succeeded!


            // All succeeded!
            {
                updateStatus(SkillStatus::Succeeded);

                // return result of main method
                std::unique_lock l(skillStatusesMutex);
                TerminatedSkillStatusUpdate ret(statusUpdate.executionId,
                                                statusUpdate.usedParameterization);
                ret.data = mainRet.data;
                ret.status = TerminatedSkillStatus::Succeeded;
                return ret;
            }
        }
    } // namespace skills::detail
} // namespace armarx
